# -*- coding: utf-8 -*-
"""
Created on Thu May  9 16:35:38 2019

@author: Dylan
"""

import datetime; now = datetime.datetime.now
from pulp import GUROBI, LpBinary, LpContinuous, LpMinimize, LpProblem
from pulp import LpVariable, lpSum
from pulp import value as pulpvalue

from MedianRoutingProblem import Solution

def apply(MRPinstance):
    """
    Calls Gurobi on this problem and creates a MRPsolution.
    Uses the MIP formulation from May 2019.
    """
    
    # Create the MIP:
    print("Constructing MIP... (%s)" % str(now()))
            
    # Initialise MILP:
    prob = LpProblem("Median Routing Problem formulation March 2019", LpMinimize)
    
    # Abbreviate sets and parameters:
    A = MRPinstance.A; J = MRPinstance.J; V = MRPinstance.V
    T = ['t%d' % i for i in range(MRPinstance.T+1)]
    Tm = T[:-1]
    VP = MRPinstance.VP; Vv = MRPinstance.Vv
    C = MRPinstance.C; P = MRPinstance.P; Q = MRPinstance.Q
    
    # Initialise variables:
    prob.x = x = LpVariable.dicts("x", (A,V ,T), 0, 1, LpBinary)
    prob.y = y = LpVariable.dicts("y", (V,VP,T), 0, 1, LpContinuous)
    prob.z = z = LpVariable.dicts("z", (A,J ,T), 0, 1, LpBinary)
    
    # Overwrite fixed values:
    for a in A:
        
        # Agent start locations:
        start_v = MRPinstance.start[a]
        for v in V:
            x[a][v][ T[0] ] = 0.0
        x[a][start_v][ T[0] ] = 1.0
        
        # Agent end locations:
        end_v = MRPinstance.end[a]
        for v in V:
            x[a][v][ T[-1] ] = 0.0
        x[a][end_v][ T[-1] ] = 1.0
    
    # Add objective function:
    prob += lpSum(P[v]*C[u][v]*y[u][v][t] for u in V for v in VP for t in T)
    
    # Add constraint (1):
#    IMPLIED BY FIXED VALUES

    # Add constraint (2):
#    IMPLIED BY FIXED VALUES
    
    # Add constraint (3):
    for a in A:
        for t in T:
            prob += lpSum( x[a][v][t] for v in V ) == 1, "(3)_%s_%s" % (a,t)

    # Add constraint (4):
    for a in A:
        for v in V:
            neighbours = Vv[v]
            for i_t, t in enumerate(Tm):
                tp1 = T[ i_t + 1 ]
                rhs = lpSum( x[a][u][t] for u in neighbours )
                prob += x[a][v][tp1] <= rhs, "(4)_%s_%s_%s" % (a,v,t)
                
    # Add constraint (5):
    for v in VP:
        for t in T:
            prob += lpSum( y[u][v][t] for u in V ) == 1, "(5)_%s_%s" % (v,t)
            
    # Add constraint (6):
    for u in V:
        for v in VP:
            for t in T:
                rhs = lpSum( x[a][u][t] for a in A )
                prob += y[u][v][t] <= rhs, "(6)_%s_%s_%s" % (u,v,t)

    # Add constraint (7):
    for j in J:
        prob += lpSum( z[a][j][t] for a in A for t in T ) == 1, "(7)_%s" % (j)
                
    # Add constraint (8):
    for a in A:
        for j in J:
            v = MRPinstance.location[j]
            
            for i_t, t in enumerate(T):
                sub_T = T[ i_t : i_t+Q[j] +1 ]
                lhs = lpSum(x[a][v][t] for t in sub_T)
                rhs = (Q[j]+1)*z[a][j][t]
                prob += lhs >= rhs, "(8)_%s_%s_%s" % (a,j,t)
                
    # Add constraint (9):
    for j in J:
        for k in J:
            if j != k and MRPinstance.location[j] == MRPinstance.location[k]:
                for a in A:
                    for i_t, t in enumerate(T):
                        subsum = lpSum(z[a][k][tau] for tau in T[i_t:i_t+Q[j]])
                        lhs = z[a][j][t] + subsum
                        prob += lhs <= 1, "(9)_%s_%s_%s_%s" % (a,j,k,t)

    # Start GUROBI:
    print("Starting GUROBI... (%s)" % str(now()))
    GUROBI().solve(prob)
    
    if prob.status != 1:
        print('Problem not solved, probably infeasible.')
        prob.writeLP('infeasiblemip.lp')
        raise ValueError
        
    # Translate result to MedianRoutingProblem.Solution:
    sol = Solution()
    sol.INSTANCE = MRPinstance
    sol.METHOD = "GUROBI, formulation March 2019"
    sol.COST = pulpvalue(prob.objective)
    sol.OPTIMAL = True
    
    sol.x_SEQUENCES = {a: list() for a in A}
    for a in A:
        for t in T:
            v = next(v for v in V if pulpvalue(x[a][v][t]))
            sol.x_SEQUENCES[a] += [v]
            
    sol.y_SEQUENCES = {v: list() for v in VP}
    for v in VP:
        for t in T:
            u = next(u for u in V if pulpvalue(y[u][v][t]))
            sol.y_SEQUENCES[v] += [u]
            
    sol.z_ASSIGNMENTS = dict()
    for j in J:
        a = next(a for a in A if sum( pulpvalue(z[a][j][t]) for t in T) )
        t = next(t for t in T if      pulpvalue(z[a][j][t])             )
        i_t = T.index(t)
        sol.z_ASSIGNMENTS[j] = (a,i_t)
        
    return sol
    