# -*- coding: utf-8 -*-
"""
Created on Tue Apr  9 14:10:10 2019

Heuristic for the basic Median Routing Problem.
Uses GUROBI to perform the following steps:
    
1. Solve k-medians (MEDIATE)
2. Get shortest distances between start, end and job locations
3. Solve 'job division' (DIVIDE)
4. For each agent, solve s-t path TSP, picking the 'clockwise' optimum (SEQUENCE)
5. For each agent, find the space-time route with best coverage to their region
6. Read out responsibility profiles (AGREE)
7. For each agent, find the space-time route with best coverage to their profile

Steps MEDIATE, SEQUENCE and AGREE can be switched off:
    - if MEDIATE is switched off, step 1 is skipped, step 2 instead solves 
        regular DVRP, and step 4 uses arbitrary fastest routes
    - if SEQUENCE is switched off, instead sequences from step 2 are used
    - if AGREE is switched off, reoptimisation steps 5 and 6 are skipped 

@author: Dylan
"""

from math import atan2
from pulp import LpBinary, LpInteger, LpMinimize, LpProblem, LpVariable, lpSum
from pulp import GUROBI, LpContinuous
from pulp import value as pulpvalue

from MedianRoutingProblem import Solution

"Default settings:"
step_Mediate_is_on = True
step_Sequence_is_on = True
step_Agree_is_on = True

def apply(MRPinstance):
    """Main function of heuristic."""
    
    if step_Mediate_is_on:
        medians, region = step1(MRPinstance)
#        print(medians)
        
    dJ_paths = step2(MRPinstance)
    
    if step_Mediate_is_on:
        clustering, init_sequence = step3(MRPinstance, medians, region, dJ_paths)
#        print(init_sequence)
        
    else:
        clustering, init_sequence = normal_DVRP(MRPinstance, dJ_paths)
        
    if step_Sequence_is_on:
        sequence = step4(MRPinstance, dJ_paths, clustering)
#        print(sequence)
        
    else:
        sequence = init_sequence
        
    if step_Mediate_is_on:
        init_route = step5(MRPinstance, dJ_paths, region, sequence)
#        print(init_route)
        
    else:
        init_route = fastest_execution(MRPinstance, dJ_paths, sequence)
    
    if step_Agree_is_on:
        profile = step6(MRPinstance, init_route)
        
        profile_route = step7(MRPinstance, dJ_paths, sequence, profile)
        
        final_route = profile_route
        
    else:
        final_route = init_route
    
#    print(final_route)
    solution = make_solution(MRPinstance, final_route)
    
    return solution

def step1(MRPinstance):
    """Solves K-MED on this instance using GUROBI.
    Returns list of |A| strings, 'medians', and a dict from agents to their
    assigned coverage regions, 'region'."""
    
    # Abrreviate:
    A = MRPinstance.A; V = MRPinstance.V; VP = MRPinstance.VP
    C = MRPinstance.C; P = MRPinstance.P
    
    # Construct the MIP for the induced k-medians:
    prob = LpProblem("kmedians", LpMinimize)
    prob.x = LpVariable.dicts("x", (V,VP), 0, 1, LpContinuous)
    prob.y = LpVariable.dicts("y", (V),    0, 1, LpBinary)
    
    # Add objective:
    prob += lpSum( P[v]*C[u][v]*prob.x[u][v] for u in V for v in VP )
    
    # Add constraints (1):
    for v in VP:
        prob += lpSum( prob.x[u][v] for u in V ) >= 1, "(1)_%s" % v
        
    # Add constraints (2):
    for u in V:
        for v in VP:
            prob += prob.x[u][v] <= prob.y[u], "(2)_%s_%s" % (u,v)
            
    # Add constraint (3):
    prob += lpSum(prob.y) <= len(A), "(3)"
    
    # Solve MIP with GUROBI:
    GUROBI().solve(prob)
    
    # Obtain the posts:
    medians = [v for v in V if pulpvalue(prob.y[v]) > 0.01]
    
    # Define the coverage region for each median/agent:
    region = {a: list() for a in A}
    for v in VP:
        m = min(medians, key = lambda m: C[m][v])
        i_m = medians.index(m)
        a = A[i_m]
        region[a] += [v]
    
    return medians, region

def step2(MRPinstance):
    """Get the shortest paths between items in dJ, where dJ is the union of
    start locations, end locations and jobs locations."""
    
    # Construct dJ:
    S = ['start_%s' % a for a in MRPinstance.A]
    E = ['end_%s' % a for a in MRPinstance.A]
    J = MRPinstance.J
    dJ = S + E + J
    
    # Store which agent belongs to the items in S and E:
    belongs = dict()
    for a in MRPinstance.A:
        belongs['start_%s' % a] = a
        belongs['end_%s' % a] = a
    
    # For the sake of speed, do not store paths only as connecting two items
    # in dJ, but also as connecting two nodes in V: this useful when, for
    # example, end locations overlap:
    dG_paths = dict()
    dJ_paths = dict()
    
    for i in dJ:
        
        # Figure out the node u where item i lives:
        if i in S:
            a = belongs[i]
            u = MRPinstance.start[a]
        elif i in E:
            a = belongs[i]
            u = MRPinstance.end[a]
        elif i in J:
            u = MRPinstance.location[i]
            
        # Prepare dicts if necessary: 
        if not i in dJ_paths:
            dJ_paths[i] = dict()
        if not u in dG_paths:
            dG_paths[u] = dict()
        
        for j in dJ:
            # Figure out the node v where item j lives:
            if j in S:
                a = belongs[j]
                v = MRPinstance.start[a]
            elif j in E:
                a = belongs[j]
                v = MRPinstance.end[a]
            elif j in J:
                v = MRPinstance.location[j]
                
            # Check if a path between u and v is already known:
            if u in dG_paths and v in dG_paths[u]:
                dJ_paths[i][j] = dG_paths[u][v]
                continue
            
            # Check if a path between v and u is already known:
            if v in dG_paths and u in dG_paths[v]:
                path = list(reversed( dG_paths[v][u] ))
                if not u in dG_paths:
                    dG_paths[u] = dict()
                dG_paths[u][v] = path
                dJ_paths[i][j] = path
                continue
            
            # Otherwise, get and store the path:
            path = shortest_path(MRPinstance, u, v)
            dG_paths[u][v] = path
            dJ_paths[i][j] = path
            
    return dJ_paths

def step3(MRPinstance, medians, region, dJ_paths):
    """Solves the 'job division' problem: using GUROBI, finds the assignment
    of jobs to the closest possible median such that each job cluster admits
    a feasible route. Returns dict of agents to their jobs, 'clustering'."""
    
    A = MRPinstance.A; J = MRPinstance.J; T = MRPinstance.T
    C = MRPinstance.C; P = MRPinstance.P; Q = MRPinstance.Q
    
    # Construct dJ:
    S = ['start_%s' % a for a in A]
    E = ['end_%s' % a for a in A]
    dJ = S + E + J
    
    # Initialise the problem, its distance and its cost function coefficients:
    prob = LpProblem('RoutingSubproblem2', LpMinimize)
    
    D = {i: {j: 0 for j in dJ} for i in dJ}
    for i in dJ_paths:
        for j in dJ_paths[i]:
            D[i][j] = len( dJ_paths[i][j] ) - 1
    
    F = {a: {j: 0.0 for j in dJ} for a in A}
    for j in J:
        u = MRPinstance.location[j]
        for a in A:
            F[a][j] = (Q[j]+1)*sum( P[v]*C[u][v] for v in region[a] )
    
    # Variables:
    prob.w = w = LpVariable.dicts('w', (dJ,dJ), 0, 1, LpBinary )
    prob.f = f = LpVariable.dicts('f', (dJ)   , 0, T, LpContinuous)
    prob.zaj = zaj = LpVariable.dicts('zaj', (A, dJ), 0, 1, LpBinary)
    
    # Set some variables fixed to 0 or 1:
    for i in dJ:
        w[i][i] = 0 # You cannot go from i to i
        if i in S:
            f[i] = 0.0 # The starting time of start points is 0
            for j in dJ:
                w[j][i] = 0 # Start points cannot be visited
        elif i in E:
            for j in dJ:
                w[i][j] = 0 # End points cannot be left
    for a in A:
        for a2 in A:
            zaj[a2][ 'start_%s' % a ] = int(a==a2) # Agents have own start
            zaj[a2][ 'end_%s' % a ] = int(a==a2) # Agents have own end
        
    # Objective:
    prob += lpSum( F[a][j]*zaj[a][j] for a in A for j in J )
    
    # Constraint (1):
    for i in S+J:
        prob += lpSum( w[i][j] for j in E+J ) == 1, "(1)_%s" % i
        
    # Constraint (2):
    for j in E+J:
        prob += lpSum( w[i][j] for i in S+J ) == 1, "(2)_%s" % j
        
    # Constraint (3):
    for i in S+J:
        q_i = Q[i] if i in J else 0
        for j in E+J:
            prob += f[j] >= f[i] + q_i + D[i][j] - T*(1-w[i][j]), "(3)_%s_%s" % (i,j)
            
    # Constraint (4):
    for a in A:
        for i in S+J:
            for j in E+J:
                prob += zaj[a][j] >= zaj[a][i] + w[i][j] - 1, "(4)_%s_%s_%s" % (a,i,j)
                
    # Constraint (5):
    for j in dJ:
        prob += lpSum(zaj[a][j] for a in A) == 1, "(5)_%s" % j
            
    # Solve:
    GUROBI().solve(prob)
    
    # Compile the clustering:
    clustering = {a: list() for a in A}
    for j in J:
        a = next(a for a in A if pulpvalue(zaj[a][j]))
        clustering[a] += [j]
    
    # Compile the initial sequencing of the jobs from the DVRP:
    init_sequence = {a: list() for a in A }
    for a in A:
        i = 'start_%s' % a
        j = next(j for j in dJ if pulpvalue(w[i][j]))
        init_sequence[a] = [i, j]
        while j != 'end_%s' % a:
            i = j
            j = next(j for j in dJ if pulpvalue(w[i][j]))
            init_sequence[a] += [j]
    
    return clustering, init_sequence

def normal_DVRP(MRPinstance, dJ_paths):
    """Just solves DVRP over the jobs, not caring about coverage regions."""

    A = MRPinstance.A; J = MRPinstance.J; T = MRPinstance.T
    
    # Construct dJ:
    S = ['start_%s' % a for a in A]
    E = ['end_%s' % a for a in A]
    dJ = S + E + J
    
    # Initialise the problem, its distance and its cost function coefficients:
    prob = LpProblem('RoutingSubproblem2', LpMinimize)
    
    D = {i: {j: 0 for j in dJ} for i in dJ}
    for i in dJ_paths:
        for j in dJ_paths[i]:
            D[i][j] = len( dJ_paths[i][j] ) - 1
    
    # Variables:
    prob.w = w = LpVariable.dicts('w', (dJ,dJ), 0, 1, LpBinary )
    prob.f = f = LpVariable.dicts('f', (dJ)   , 0, T, LpContinuous)
    prob.zaj = zaj = LpVariable.dicts('zaj', (A, dJ), 0, 1, LpBinary)
    
    # Set some variables fixed to 0 or 1:
    for i in dJ:
        w[i][i] = 0 # You cannot go from i to i
        if i in S:
            f[i] = 0.0 # The starting time of start points is 0
            for j in dJ:
                w[j][i] = 0 # Start points cannot be visited
        elif i in E:
            for j in dJ:
                w[i][j] = 0 # End points cannot be left
    for a in A:
        for a2 in A:
            zaj[a2][ 'start_%s' % a ] = int(a==a2) # Agents have own start
            zaj[a2][ 'end_%s' % a ] = int(a==a2) # Agents have own end
        
    # Objective:
    prob += lpSum( D[i][j]*w[i][j] for i in dJ for j in dJ )
    
    # Constraint (1):
    for i in S+J:
        prob += lpSum( w[i][j] for j in E+J ) == 1, "(1)_%s" % i
        
    # Constraint (2):
    for j in E+J:
        prob += lpSum( w[i][j] for i in S+J ) == 1, "(2)_%s" % j
        
    # Constraint (3):
    for i in S+J:
        q_i = MRPinstance.Q[i] if i in J else 0
        for j in E+J:
            prob += f[j] >= f[i] + q_i + D[i][j] - T*(1-w[i][j]), "(3)_%s_%s" % (i,j)
            
    # Constraint (4):
    for a in A:
        for i in S+J:
            for j in E+J:
                prob += zaj[a][j] >= zaj[a][i] + w[i][j] - 1, "(4)_%s_%s_%s" % (a,i,j)
                
    # Constraint (5):
    for j in dJ:
        prob += lpSum(zaj[a][j] for a in A) == 1, "(5)_%s" % j
            
    # Solve (GUROBI):
    GUROBI().solve(prob)
    
    # Compile the clustering:
    clustering = {a: list() for a in A}
    for j in J:
        a = next(a for a in A if pulpvalue(zaj[a][j]))
        clustering[a] += [j]
    
    # Compile the initial sequencing of the jobs from the DVRP:
    init_sequence = {a: list() for a in A }
    for a in A:
        i = 'start_%s' % a
        j = next(j for j in dJ if pulpvalue(w[i][j]))
        init_sequence[a] = [i, j]
        while j != 'end_%s' % a:
            i = j
            j = next(j for j in dJ if pulpvalue(w[i][j]))
            init_sequence[a] += [j]
    
    return clustering, init_sequence
    

def step4(MRPinstance, dJ_paths, clustering):
    """For each agent, solves the s-t path TSP. Then if the start point and 
    end point are the same, orients the sequence so it's 'clockwise'.
    Returns a dict from agents to their sequence of jobs."""
    
    # Abbreviate:
    A = MRPinstance.A; T = MRPinstance.T
    Q = MRPinstance.Q; crd = MRPinstance.coord
    
    sequence = {a: None for a in A}
    
    for a in A:
        
        # If a has no jobs, its sequence simply exists of starting and ending
        # at its start and end location:
        if len(clustering[a]) == 0:
            sequence[a] = ['start_%s' % a, 'end_%s' % a]
            continue
        
        # Otherwise, get the distances between the goals of a:
        Sa = ['start_%s' % a]
        Ea = ['end_%s' % a]
        Ja = clustering[a]
        dJa = Sa + Ea + Ja
        D = {i: {j: len(dJ_paths[i][j]) - 1 for j in dJa} for i in dJa}
                
        # Solve TSP on these:
        prob = LpProblem('s-t-TSP', LpMinimize)
        prob.x = x = LpVariable.dicts('x', (dJa, dJa), 0, 1, LpBinary)
        prob.u = u = LpVariable.dicts('u', (dJa), 0, T, LpInteger)
        for i in dJa:
            x[i][i] = 0 # You cannot go from i to i
        u['start_%s' % a] = 0 # The start time at the start is 0
        for i in dJa:
            x[i]['start_%s' % a] = 0 # You cannot go to the start
            x['end_%s' % a][i] = 0 # You cannot leave from the end
        
        # Objective:
        prob += lpSum( D[i][j]*x[i][j] for i in dJa for j in dJa )
        
        # Constraint (1):
        for i in Sa+Ja:
            prob += lpSum( x[i][j] for j in Ea+Ja ) == 1, "(1)_%s" % (i)
            
        # Constraint (2):
        for j in Ea+Ja:
            prob += lpSum( x[i][j] for i in Sa+Ja ) == 1, "(2)_%s" % (j)
        
        # Constraint (3):
        for i in Sa+Ja:
            q = Q[i] if i in Ja else 0
            for j in Ea+Ja:
                if i != j:
                    prob += u[j] >= u[i] + q + D[i][j] - T*(1-x[i][j]), "(3)_%s_%s" % (i,j)
        
        # Solve:            
        GUROBI().solve(prob)
        
        # Read out the sequence of a:
        i = 'start_%s' % a
        j = next(j for j in dJa if pulpvalue(x[i][j]))
        sequence_a = [i, j]
        while j != 'end_%s' % a:
            i = j
            j = next(j for j in dJa if pulpvalue(x[i][j]))
            sequence_a += [j]

        # If start[a] = end[a] and the sequence of a is counter-clockwise, 
        # reverse it, but don't have the sequence start with end_a:
        if MRPinstance.start[a] == MRPinstance.end[a] and len(sequence_a) > 3:
            depot = MRPinstance.start[a]
            nr_clockwise_moves = 0
            for i_i, i in enumerate(sequence_a[:-1]):
                j = sequence_a[i_i+1]
                u = MRPinstance.location[i] if i in Ja else MRPinstance.start[a]
                v = MRPinstance.location[j] if j in Ja else MRPinstance.end[a]
                angle_u = atan2(crd[u][1] - crd[depot][1],
                                crd[u][0] - crd[depot][0])
                angle_v = atan2(crd[v][1] - crd[depot][1],
                                crd[v][0] - crd[depot][0])
                if angle_u > angle_v:
                    nr_clockwise_moves += 1
                else:
                    nr_clockwise_moves -= 1
            print(nr_clockwise_moves) #debgu
            if nr_clockwise_moves < 0:
                sequence_a = list(reversed(sequence_a))
                sequence_a[0] = 'start_%s' % a
                sequence_a[-1] = 'end_%s' % a
            
        sequence[a] = sequence_a
        
    return sequence

def step5(MRPinstance, dJ_paths, region, sequence):
    """For each agent, execute their sequence with the best possible coverage
    of the assigned region. Returns a dict from agents to space-time paths."""
    
    A = MRPinstance.A 
    set_T = ['t%d' % i_t for i_t in range( MRPinstance.T+1 )]
    
    profile = {a: {t: region[a] for t in set_T} for a in A}
    region_route = execute_sequence(MRPinstance, dJ_paths, sequence, profile)
        
    return region_route

def fastest_execution(MRPinstance, dJ_paths, sequence):
    """For each agent, execute their sequence caring only about being back
    at the depot as quickly as possible."""
    
    A = MRPinstance.A; J = MRPinstance.J
    set_T = [ 't%d' % i_t for i_t in range(MRPinstance.T+1) ]
    start = MRPinstance.start; end = MRPinstance.end
    Q = MRPinstance.Q; location = MRPinstance.location
    
    fastest_route = {a: list() for a in A}
    
    for a in A:
        
        # Read out x_SEQUENCE:
        x_SEQUENCE = [ start[a] ]
        for i_j, j in enumerate( sequence[a][1:] ):
            i = sequence[a][i_j]
            x_SEQUENCE += dJ_paths[i][j][1:]
            if j in J:
                x_SEQUENCE += Q[j]*[ location[j] ]
                
        if len(x_SEQUENCE) < MRPinstance.T+1:
            difference = MRPinstance.T+1 - len(x_SEQUENCE)
            x_SEQUENCE += difference*[ end[a] ]
            
        # Knowing this, construct its equivalent (v,t,state)-representation:
        state = sequence[a][1]
        for i_t, t in enumerate(set_T):
            v = x_SEQUENCE[i_t]
            
            if state in J and v == location[state] and Q[state] > 0:
                new_state = ( state, Q[state] )
                
            elif state in J and v == location[state] and Q[state] == 0:
                i_state = sequence[a].index(state)
                new_state = sequence[a][ i_state+1 ]
                
            elif isinstance(state, tuple) and state[1] > 1:
                new_state = (state[0], state[1] - 1)
                
            elif isinstance(state, tuple) and state[1] == 1:
                i_goal = sequence[a].index(state[0])
                new_state = sequence[a][ i_goal+1 ]
                
            else:
                new_state = state
                
            fastest_route[a] += [ (v,t,new_state) ]
            state = new_state
                
    return fastest_route

def step6(MRPinstance, region_route):
    """Simply checks for each node and each time-step which agent is closest
    according to region_route. Returns a dict from agents and time-steps to 
    their assigned space-time nodes."""
    
    A = MRPinstance.A; VP = MRPinstance.VP
    set_T = ['t%d' % i_t for i_t in range( MRPinstance.T + 1 )]
    C = MRPinstance.C
    
    profile = {a: {t: list() for t in set_T} for a in A}
    for i_t, t in enumerate(set_T):
        for v in VP:
            a = min(A, key = lambda a: C[ region_route[a][ i_t ][0] ][v])
            profile[a][t] += [v]
            
    return profile

def step7(MRPproblem, dJ_paths, sequence, profile):
    """Same as step 4, except now coverage is given to a dynamic responsibility
    profile."""
        
    profile_route = execute_sequence(MRPproblem, dJ_paths, sequence, profile)
    
    return profile_route

def make_solution(MRPinstance, route):
    """Given any route, translates it to a solution to the given MRP."""
    
    A = MRPinstance.A; J = MRPinstance.J; VP = MRPinstance.VP
    C = MRPinstance.C; P = MRPinstance.P; Q = MRPinstance.Q
    
    # Read out x_SEQUENCES:
    x_SEQUENCES = {a: list() for a in A}
    for a in A:
        for (v,t,state) in route[a]:
            x_SEQUENCES[a] += [v]
            
    # Infer y_SEQUENCES and increment COST:            
    y_SEQUENCES = {v: list() for v in VP}
    COST = 0
    for i_t in range( MRPinstance.T+1 ):
        U = [ x_SEQUENCES[a][i_t] for a in A ]
        for v in VP:
            u = min(U, key = lambda u: C[u][v])
            y_SEQUENCES[v] += [u]
            COST += C[u][v]*P[v]
            
    # Read out z_ASSIGNMENTS:
    z_ASSIGNMENTS = {j: None for j in J}
    for a in A:
        for i_3node, (v,t,state) in enumerate(route[a]):
            if isinstance(state, tuple):
                (j, k) = state
                if k == Q[j]:
                    z_ASSIGNMENTS[j] = (a, i_3node)
            elif state in J and Q[state] == 0:
                (vp1, tp1, statep1) = route[a][ i_3node +1 ]
                if statep1 != state:
                    z_ASSIGNMENTS[state] = (a, i_3node +1)
    
    # Build METHOD name:
    METHOD = ""
    if step_Mediate_is_on:
        METHOD += "M"
    METHOD += "D"
    if step_Sequence_is_on:
        METHOD += "S"
    if step_Agree_is_on:
        METHOD += "A"
    METHOD += ", formulation April 2019"
    
    # Translate result to MedianRoutingProblem.Solution and return:
    sol = Solution()
    sol.INSTANCE = MRPinstance
    sol.METHOD = "Wait At Medians, formulation April 2019"
    sol.COST = COST
    sol.OPTIMAL = False
    sol.x_SEQUENCES = x_SEQUENCES
    sol.y_SEQUENCES = y_SEQUENCES
    sol.z_ASSIGNMENTS = z_ASSIGNMENTS
    
    return sol

def execute_sequence(MRPinstance, dJ_paths, sequence, profile):
    
    """Given an agent a, the sequence of jobs they visit and the space-time
    nodes they carry responsibility over, compute the cheapest space-time path
    over the sequence with respect to expected response time."""
    
    # Strategy: Dijkstra, where nodes are three-dimensional: 
    # (V, T, state (either next goal or processing j with i timesteps left))
    # The cost incurred by visiting a node is the expected response time from
    # that node. The cost of traversing an arc is equal to the cost of the
    # visited node: therefore, the Dijkstra cost of that node never changes
    # after the first discovery. Therefore, the optimal x is found as soon as
    # the final node, (depot, T[-1], hyperroute[-1]), is reached.
    
    orp = MRPinstance
    A = orp.A; J = orp.J
    Vv = orp.Vv
    set_T = ['t%d' % i_t for i_t in range( orp.T+1 )]

    C = orp.C; P = orp.P; Q = orp.Q
    
    region_route = {a: None for a in A}
    
    # Denote D_ij the distance between i and j, thus 
    # the length of dJ_paths[i][j] minus 1:
    D = {i: {j: len(dJ_paths[i][j])-1 for j in dJ_paths[i]} for i in dJ_paths}
    
    for a in A:
        
        hyperroute = sequence[a]
        end_item = 'end_%s' % a
        
        # Determine deadlines of the jobs, implied by having to be back at the
        # depot in time after processing all subsequent jobs:
        deadline = {j: None for j in J+[end_item]}
        for j in reversed(hyperroute[1:]):
            if j == end_item:
                deadline[j] = set_T[-1]
            else:
                i_j = hyperroute.index(j)
                jp1 = hyperroute[i_j+1]
                
                g = D[j][jp1]
                q = Q[j]
                
                # Deadline is deadline of the next job, minus processing time of
                # this job and minimum travel time:
                tp1 = deadline[jp1]
                i_tp1 = set_T.index(tp1)
                i_t = i_tp1 - g - q
                deadline[j] = set_T[i_t]
                
        # Start Dijkstra:
        start = (orp.start[a], set_T[0], hyperroute[1])
        end = (orp.end[a], set_T[-1], end_item)
        
        node_cost = dict()
        dijkstra_cost = {start: sum( P[w]*C[ orp.start[a] ][w] 
                                            for w in profile[a][ set_T[0] ] )}
        active_nodes = [start]
        explored_nodes = list()
        
        while not end in dijkstra_cost:
            
            cur_node = min(active_nodes, key = lambda n: dijkstra_cost[n])
            (u, t, state) = cur_node
            active_nodes.remove(cur_node)
            explored_nodes += [cur_node]
            
            # If t == T[-1], this node is hopeless:
            if t == set_T[-1]:
                continue
            tp1 = set_T[ set_T.index(t)+1 ]
            
            # Find/define neighbours in space-time, disqualifying those that 
            # lead to violated deadlines:
            neighbours = list()
            
            if state == end_item:
                for v in Vv[u]:
                    neighbours += [ (v, tp1, end_item) ]
            
            if state in J and Q[state] > 0:
                for v in Vv[u]:
                    if deadline[state] != tp1:
                        neighbours += [ (v, tp1, state) ]
                    if v == orp.location[state]:
                        neighbours += [ (v, tp1, (state, Q[state]) )]
                        
            if state in J and Q[state] == 0:
                for v in Vv[u]:
                    if deadline[state] != tp1 and v != orp.location[state]:
                        neighbours += [ (v, tp1, state) ]
                    if v == orp.location[state]:
                        i_j = hyperroute.index(state)
                        jp1 = hyperroute[ i_j+1 ]
                        neighbours += [ (v, tp1, jp1) ]
                    
            if isinstance(state, tuple) and state[1] > 1:
                neighbours += [ (u, tp1, (state[0], state[1]-1) ) ]
                
            if isinstance(state, tuple) and state[1] == 1:
                i_j = hyperroute.index(state[0])
                jp1 = hyperroute[ i_j+1 ]
                if tp1 != deadline[jp1]:
                    neighbours += [ (u, tp1, jp1) ]
                if jp1 in J and orp.location[jp1] == u and Q[jp1] > 0:
                    neighbours += [ (u, tp1, (jp1, Q[jp1]) ) ]
                                    
            # Now, explore neighbours. If they were previously undiscovered, set
            # them to active and chart down their Dijkstra cost; otherwise, their
            # cost cannot be improved and ignore them:
            for new_node in neighbours:
                (v, tp1, new_state) = new_node
                
                if not new_node in dijkstra_cost:
                    node_cost[new_node] = sum( P[w]*C[v][w] for w in profile[a][tp1] )
                    dijkstra_cost[ new_node ] = dijkstra_cost[ cur_node ] + node_cost[new_node]
                    active_nodes += [ new_node ]
                
        # Reconstruct the path found this way:
        path = [ end ]
        while not start in path:
            (v, tp1, new_state) = v_node = path[0]
            
            if isinstance(new_state, tuple):
                (j, k) = new_state
                if k == Q[j]:
                    candidates = [(u, t, state) for (u, t, state) in explored_nodes
                                  if v in Vv[u] 
                                  and set_T.index(t) == set_T.index(tp1) - 1
                                  and (state == j or (isinstance(state, tuple)
                                  and  state[1] == 1 and u == v))]
                    
                else:
                    candidates = [(u, t, state) for (u, t, state) in explored_nodes
                                  if v in Vv[u]
                                  and set_T.index(t) == set_T.index(tp1) - 1
                                  and state == (j, k+1)]
            else:
                candidates = [(u, t, state) for (u, t, state) in explored_nodes
                              if v in Vv[u]
                              and set_T.index(t) == set_T.index(tp1) - 1
                              and state == new_state] 

                candidates += [(u, t, state) for (u, t, state) in explored_nodes
                               if u == v
                               and set_T.index(t) == set_T.index(tp1) - 1
                               and (isinstance(state, tuple) and state[1] == 1)]
                
                candidates += [(u, t, state) for (u, t, state) in explored_nodes
                               if v in Vv[u]
                               and set_T.index(t) == set_T.index(tp1) - 1
                               and (state in J and Q[state] == 0 and
                                    hyperroute[ hyperroute.index(state)+1 ] == new_state)]
                              
            prev_node = next( u_node for u_node in candidates 
                             if abs(dijkstra_cost[u_node] + node_cost[v_node] - dijkstra_cost[v_node]) < 0.001 )
            path = [prev_node] + path
            
        region_route[a] = path
        
    return region_route

def shortest_path(MRPinstance, u, v):
    """Uses width-first search to find a shortest path from u to v."""
    
    shortest = {u: [u]}
    active_nodes = [u]
    
    while not v in shortest:
        
        # Pick an unexplored node:
        w = active_nodes[0]
        active_nodes = active_nodes[1:]
        
        # Check its neighbours: new neighbours get stored in shortest and
        # in active_nodes:
        for neighbour in MRPinstance.Vv[w]:
            if not neighbour in shortest:
                shortest[ neighbour ] = shortest[ w ] + [ neighbour ]
                active_nodes += [ neighbour ]
                
                if neighbour == v:
                    break
            
    if not v in shortest:
        raise ValueError
        
    else:
        return shortest[v]