# -*- coding: utf-8 -*-
"""
Created on Mon Apr  8 10:33:53 2019

@author: Dylan
"""

import datetime; now = datetime.datetime.now
from pulp import lpSum, LpBinary, LpContinuous, LpMinimize, LpProblem
from pulp import LpVariable, GUROBI
from pulp import value as pulpvalue

from MedianRoutingProblem import Solution

def apply(MRPinstance):
    """
    Uses the Wait-At-Medians heuristic to construct a MRPsolution.
    Uses the WAM formulation from April 2019.
    """
    
    "1. Solve K-MED:"
    medians = step1(MRPinstance)
    
    "2. Get shortest paths between dJ:"
    paths = step2(MRPinstance, medians)
    
    "3. Solve Routing Subproblem 1:"
    w_values = step3(MRPinstance, medians, paths)
    
    "4. Translate back to discrete setting:"
    solution = step4(MRPinstance, medians, paths, w_values)
    
    return solution

def step1(MRPinstance):
    """Solve K-MED: solve the k-medians problem on this instance, where k
    is the amount of agents."""
    
    # Abrreviate:
    A = MRPinstance.A; V = MRPinstance.V; VP = MRPinstance.VP
    C = MRPinstance.C; P = MRPinstance.P
    
    # Construct the MIP for the induced k-medians:
    prob = LpProblem("kmedians", LpMinimize)
    prob.x = LpVariable.dicts("x", (V,VP), 0, 1, LpContinuous)
    prob.y = LpVariable.dicts("y", (V),    0, 1, LpBinary)
    
    # Add objective:
    prob += lpSum( P[v]*C[u][v]*prob.x[u][v] for u in V for v in VP )
    
    # Add constraints (1):
    for v in VP:
        prob += lpSum( prob.x[u][v] for u in V ) >= 1, "(1)_%s" % v
        
    # Add constraints (2):
    for u in V:
        for v in VP:
            prob += prob.x[u][v] <= prob.y[u], "(2)_%s_%s" % (u,v)
            
    # Add constraint (3):
    prob += lpSum(prob.y) <= len(A), "(3)"
    
    # Solve MIP with GUROBI:
    GUROBI().solve(prob)
    
    # Obtain the posts:
    medians = [v for v in V if pulpvalue(prob.y[v]) > 0.01]
    
    return medians

def step2(MRPinstance, medians):
    """Get the shortest paths between items in dJ, where dJ is the union of
    start locations, end locations, jobs and medians."""
    
    # Construct dJ:
    S = ['start_%s' % a for a in MRPinstance.A]
    M = medians
    E = ['end_%s' % a for a in MRPinstance.A]
    J = MRPinstance.J
    dJ = S + M + E + J
    
    # Store which agent belongs to the items in S and E:
    belongs = dict()
    for a in MRPinstance.A:
        belongs['start_%s' % a] = a
        belongs['end_%s' % a] = a
    
    # For the sake of speed, do not store paths only as connecting two items
    # in dJ, but also as connecting two nodes in V: this useful when, for
    # example, end locations overlap:
    dG_paths = dict()
    dJ_paths = dict()
    
    for i in dJ:
        
        # Figure out the node u where item i lives:
        if i in S:
            a = belongs[i]
            u = MRPinstance.start[a]
        elif i in E:
            a = belongs[i]
            u = MRPinstance.end[a]
        elif i in J:
            u = MRPinstance.location[i]
        elif i in M:
            u = i
            
        # Prepare dicts if necessary: 
        if not i in dJ_paths:
            dJ_paths[i] = dict()
        if not u in dG_paths:
            dG_paths[u] = dict()
        
        for j in dJ:
            # Figure out the node v where item j lives:
            if j in S:
                a = belongs[j]
                v = MRPinstance.start[a]
            elif j in E:
                a = belongs[j]
                v = MRPinstance.end[a]
            elif j in J:
                v = MRPinstance.location[j]
            elif j in M:
                v = j
                
            # Check if a path between u and v is already known:
            if u in dG_paths and v in dG_paths[u]:
                dJ_paths[i][j] = dG_paths[u][v]
                continue
            
            # Check if a path between v and u is already known:
            if v in dG_paths and u in dG_paths[v]:
                path = list(reversed( dG_paths[v][u] ))
                if not u in dG_paths:
                    dG_paths[u] = dict()
                dG_paths[u][v] = path
                dJ_paths[i][j] = path
                continue
            
            # Otherwise, get and store the path:
            path = shortest_path(MRPinstance, u, v)
            dG_paths[u][v] = path
            dJ_paths[i][j] = path
            
    return dJ_paths
    
def step3(MRPinstance, medians, paths):
    """Solve Routing Subproblem 1: let all agents start at their start, visit
    exactly one median of their own and end at their end, such that all jobs 
    are visited and everyone is at their end before T and the travelled 
    distance is minimal."""

    A = MRPinstance.A; J = MRPinstance.J; T = MRPinstance.T
    
    # Construct dJ:
    S = ['start_%s' % a for a in A]
    M = medians
    E = ['end_%s' % a for a in A]
    dJ = S + M + E + J
    
    # Initialise the problem and its cost function coefficients:
    prob = LpProblem('RoutingSubproblem1', LpMinimize)
    D = {i: {j: 0 for j in dJ} for i in dJ}
    for i in paths:
        for j in paths[i]:
            D[i][j] = len( paths[i][j] ) - 1
    
    # Variables:
    prob.w = w = LpVariable.dicts('w', (dJ,dJ), 0, 1, LpBinary )
    prob.f = f = LpVariable.dicts('f', (dJ)   , 0, T, LpContinuous)
    prob.zaj = zaj = LpVariable.dicts('zaj', (A, dJ), 0, 1, LpBinary)
    
    # Set some variables fixed to 0 or 1:
    for i in dJ:
        w[i][i] = 0 # You cannot go from i to i
        if i in S:
            f[i] = 0.0 # The starting time of start points is 0
            for j in dJ:
                w[j][i] = 0 # Start points cannot be visited
        elif i in E:
            for j in dJ:
                w[i][j] = 0 # End points cannot be left
    for a in A:
        for a2 in A:
            zaj[a2][ 'start_%s' % a ] = int(a==a2) # Agents have own start
            zaj[a2][ 'end_%s' % a ] = int(a==a2) # Agents have own end
        
    # Objective:
    prob += lpSum( D[i][j]*w[i][j] for i in dJ for j in dJ )
    
    # Constraint (1):
    for i in S+M+J:
        prob += lpSum( w[i][j] for j in M+E+J ) == 1, "(1)_%s" % i
        
    # Constraint (2):
    for j in M+E+J:
        prob += lpSum( w[i][j] for i in S+M+J ) == 1, "(2)_%s" % j
        
    # Constraint (3):
    for i in S+M+J:
        q_i = MRPinstance.Q[i] if i in J else 0
        for j in M+E+J:
            prob += f[j] >= f[i] + q_i + D[i][j] - T*(1-w[i][j]), "(3)_%s_%s" % (i,j)
            
    # Constraint (4):
    for a in A:
        for i in S+M+J:
            for j in M+E+J:
                prob += zaj[a][j] >= zaj[a][i] + w[i][j] - 1, "(4)_%s_%s_%s" % (a,i,j)
                
    # Constraint (5):
    for j in dJ:
        prob += lpSum(zaj[a][j] for a in A) == 1, "(5)_%s" % j
        
    # Constraint (6):
    for a in A:
        prob += lpSum(zaj[a][m] for m in M) == 1, "(6)_%s" % a
        
    # Solve and read out w_values:
    GUROBI().solve(prob)
    w_values = {i: {j: pulpvalue(w[i][j]) for j in dJ} for i in dJ}
    
    return w_values

def step4(MRPinstance, medians, paths, w_values):
    """Translate back to discrete setting."""

    A = MRPinstance.A; J = MRPinstance.J

    # Construct dJ:
    S = ['start_%s' % a for a in A]
    M = medians
    E = ['end_%s' % a for a in A]
    dJ = S + M + E + J
    
    # For each agent, compose their x-SEQUENCE:
    x_SEQUENCES = {a: list() for a in A}
    goalpath = {a: None for a in A}
    for a in A:
        
        # First compose their path over goals:
        i = 'start_%s' % a
        j = next(j for j in dJ if w_values[i][j])
        goalpath[a] = [i, j]
        while j != 'end_%s' % a:
            i = j
            j = next(j for j in dJ if w_values[i][j])
            goalpath[a] += [j]
            
        # Next, compute how much time would be left if no waiting were done:
        timeleft = MRPinstance.T
        for i_i, i in enumerate(goalpath[a][:-1]):
            j = goalpath[a][ i_i+1 ]
            timeleft -= ( len(paths[i][j]) - 1 )
            if i in J:
                timeleft -= MRPinstance.Q[i]
        
        # Read out the paths from step 2 and wait only at the median:
        x_SEQUENCES[a] = [ MRPinstance.start[a] ]
        for i_i, i in enumerate(goalpath[a][:-1]):
            j = next(j for j in dJ if w_values[i][j])
            x_SEQUENCES[a] += paths[i][j][1:]
            if j in J:
                x_SEQUENCES[a] += [ MRPinstance.location[j] ]*MRPinstance.Q[j]
            elif j in M:
                x_SEQUENCES[a] += [j]*timeleft
                
    # For each emergency node, compose its y-SEQUENCE and increment COST:
    y_SEQUENCES = {v: list() for v in MRPinstance.VP}
    COST = 0
    for i_t in range( MRPinstance.T+1 ):
        U = [ x_SEQUENCES[a][i_t] for a in A ]
        for v in MRPinstance.VP:
            u = min(U, key = lambda u: MRPinstance.C[u][v])
            y_SEQUENCES[v] += [u]
            COST += MRPinstance.C[u][v]*MRPinstance.P[v]
    
    # Get job assignments and times by checking the path of each agent:
    z_ASSIGNMENTS = {j: None for j in J}
    for a in A:
        i_t = 0
        for j in goalpath[a][:-1]:
            if j in J:
                v = MRPinstance.location[j]
                i_v = x_SEQUENCES[a][i_t:].index(v)
                i_t += i_v
                z_ASSIGNMENTS[j] = (a, i_t)
                i_t += MRPinstance.Q[j] + 1
                
    # Translate result to MedianRoutingProblem.Solution and return:
    sol = Solution()
    sol.INSTANCE = MRPinstance
    sol.METHOD = "Wait At Medians, formulation April 2019"
    sol.COST = COST
    sol.OPTIMAL = False
    sol.x_SEQUENCES = x_SEQUENCES
    sol.y_SEQUENCES = y_SEQUENCES
    sol.z_ASSIGNMENTS = z_ASSIGNMENTS
    
    return sol
        
def shortest_path(MRPinstance, u, v):
    """Uses width-first search to find a shortest path from u to v."""
    
    shortest = {u: [u]}
    active_nodes = [u]
    
    while not v in shortest:
        
        # Pick an unexplored node:
        w = active_nodes[0]
        active_nodes = active_nodes[1:]
        
        # Check its neighbours: new neighbours get stored in shortest and
        # in active_nodes:
        for neighbour in MRPinstance.Vv[w]:
            if not neighbour in shortest:
                shortest[ neighbour ] = shortest[ w ] + [ neighbour ]
                active_nodes += [ neighbour ]
                
                if neighbour == v:
                    break
            
    if not v in shortest:
        raise ValueError
        
    else:
        return shortest[v]
