# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 12:57:09 2019

@author: Dylan
"""

from importlib import import_module
from networkx import DiGraph, Graph, draw_networkx_edges, draw_networkx_nodes
from traceback import print_exc

# Default parameters:
sections = ['NODE_COORD', 'ADJACENCY', 'AGENT_START_LOCATION',
            'AGENT_END_LOCATION', 'JOB_LOCATION', 'JOB_DURATION',
            'EMERGENCY_PROBABILITY', 'RESPONSE_TIME']
default_solve_method = 'MRP_gurobi_3_19'
epsilon = 0.0001

class Instance:
    """An instance of the Median Routing Problem. """

    def __init__(self, **kwargs):

        self.NAME = None
        self.COMMENT = None
        self.TYPE = None
        self.AMOUNT_OF_NODES = None
        self.AMOUNT_OF_EMERGENCY_NODES = None
        self.NODE_ORDER = None
        self.ADJACENCY_TYPE = None
        self.ADJACENCY_FORMAT = None
        self.RESPONSE_TIME_TYPE = None
        self.RESPONSE_TIME_FORMAT = None

        self.coord = None
        self.start = None
        self.end = None
        self.location = None

        self.A = None
        self.J = None
        self.V = None
        self.VP = None
        self.Vv = None
        self.T = None
        self.C = None
        self.Q = None
        self.P = None

        for arg in kwargs:
            self.__dict__[arg] = kwargs[arg]

    def from_txt(filename):
        """Read an instance from a properly formatted .txt-file."""

        try:
            # Parse file:
            fw = open(filename)
            lines = fw.readlines()
            fw.close()

            # Initialise object:
            mrp = Instance()
            mrp.A = list()
            mrp.J = list()
            mrp.V = list()
            mrp.VP = list()
            mrp.C = dict()
            mrp.P = dict()
            mrp.Q = dict()
            mrp.Vv = dict()
            mrp.coord = dict()
            mrp.start = dict()
            mrp.end = dict()
            mrp.location = dict()

            # Analyse lines.
            # If a line contains a :, that means it's an attribute that can be
            # directly attached to mrp.
            # If a line contains "_SECTION", it means a new section is starting.
            # If it has neither, the line is part of the current section.
            expecting = {section: False for section in sections}
            awaiting_identifier = True
            response_time_i_u = 0
            response_time_i_v = 0
            response_time_triangle_mode = False
            for line in lines:

                if ':' in line:

                    # Get name of attribute:
                    splitline = line.split(':')
                    arg = splitline[0]

                    # Get value after the : and turn it into int or float if
                    # possible:
                    val = line[ len(arg)+2: ].split('\n')[0]
                    try:
                        val = int(val)
                    except ValueError:
                        try:
                            val = float(val)
                        except ValueError:
                            pass

                    # Attach attribute:
                    mrp.__dict__[arg] = val

                    # For some args, immediately enact something:
                    if arg == 'AMOUNT_OF_NODES':
                        mrp.V = [ 'v%d' % (i) for i in range(val) ]
                        mrp.Vv = { v: {v} for v in mrp.V }
                        mrp.C = { u: dict() for u in mrp.V }
                        
                    elif arg == 'AMOUNT_OF_EMERGENCY_NODES':
                        first_emergency_node = (mrp.AMOUNT_OF_NODES - 
                                                mrp.AMOUNT_OF_EMERGENCY_NODES
                                                + 1)
                        mrp.VP = mrp.V[ first_emergency_node-1 : ]
                        
                        if mrp.AMOUNT_OF_NODES == mrp.AMOUNT_OF_EMERGENCY_NODES:
                            response_time_i_u = 1
                            response_time_triangle_mode = True

                elif 'SECTION' in line:
                    expecting = {section: False for section in sections}
                    section = "_".join(line.split('_')[:-1])
                    expecting[section] = True

                elif 'EOF' in line:
                    pass

                else:

                    # Coordinates of nodes:
                    if expecting['NODE_COORD']:

                        entries = line.split('\n')[0].split('\t')
                        i_v = int(entries[0]) - 1
                        v = 'v%d' % i_v
                        coord = ( float(entries[1]), float(entries[2]) )
                        mrp.coord[v] = coord

                    # Adjacency:
                    if expecting['ADJACENCY']:

                        if line == '\n':
                            awaiting_identifier = True

                        else:
                            entries = line.split('\n')[0].split('\t')

                            first_col = 0

                            if awaiting_identifier:
                                i_u = int(entries[0]) - 1
                                u = 'v%d' % i_u
                                awaiting_identifier = False
                                first_col = 1

                            for entry in entries[first_col:]:
                                if entry != '':
                                    i_v = int(entry) - 1
                                    v = 'v%d' % i_v
                                    mrp.Vv[u].add(v)
                                    mrp.Vv[v].add(u)

                    # Agent start locations:
                    if expecting['AGENT_START_LOCATION']:

                        entries = line.split('\n')[0].split('\t')
                        i_a = int(entries[0]) - 1
                        a = 'a%d' % i_a
                        i_u = int(entries[1]) - 1
                        u = 'v%d' % i_u
                        mrp.start[a] = u

                    # Agent end locations:
                    if expecting['AGENT_END_LOCATION']:

                        entries = line.split('\n')[0].split('\t')
                        i_a = int(entries[0]) - 1
                        a = 'a%d' % i_a
                        i_v = int(entries[1]) - 1
                        v = 'v%d' % i_v
                        mrp.end[a] = v

                    # Job locations:
                    if expecting['JOB_LOCATION']:

                        entries = line.split('\n')[0].split('\t')
                        i_j = int(entries[0]) - 1
                        j = 'j%d' % i_j
                        i_v = int(entries[1]) - 1
                        v = 'v%d' % i_v
                        mrp.location[j] = v

                    # Job durations:
                    if expecting['JOB_DURATION']:

                        entries = line.split('\n')[0].split('\t')
                        i_j = int(entries[0]) - 1
                        j = 'j%d' % i_j
                        q = int(entries[1])
                        mrp.Q[j] = q

                    # Emergency probabilities:
                    if expecting['EMERGENCY_PROBABILITY']:

                        entries = line.split('\n')[0].split('\t')
                        i_v = int(entries[0]) - 1
                        v = 'v%d' % i_v
                        p = float(entries[1])
                        mrp.P[v] = p

                    # Response time, following the RESPONSE_TIME_FORMAT:
                    if expecting['RESPONSE_TIME']:

                        entries = line.split('\n')[0].split('\t')

                        for entry in entries:
                            if entry != '':
                                u = 'v%d' % response_time_i_u
                                v = mrp.VP[response_time_i_v]
                                mrp.C[u][v] = float(entry)
    
                                if not response_time_triangle_mode:
                                    response_time_i_v += 1
                                    if response_time_i_v >= mrp.AMOUNT_OF_EMERGENCY_NODES:
                                        response_time_i_u += 1
                                        response_time_i_v = 0
    
                                        if response_time_i_u == first_emergency_node - 1:
                                            response_time_triangle_mode = True
                                            response_time_i_u += 1
    
                                else:
                                    mrp.C[v][u] = float(entry)
    
                                    response_time_i_v += 1
                                    if response_time_i_v >= response_time_i_u - first_emergency_node + 1:
                                        response_time_i_u += 1
                                        response_time_i_v = 0

            # For v in VP, also set C[v][v] = 0.0:
            for v in mrp.VP:
                mrp.C[v][v] = 0.0

            # Also make A and J, then return:
            mrp.A = [ 'a%d' % (i_a) for i_a in range( len(mrp.start) ) ]
            mrp.J = [ 'j%d' % (i_j) for i_j in range( len(mrp.location) ) ]

            return mrp

        except:
            print("Could not read instance from txt.")
            print_exc()

    def to_txt(self, filename):
        """Write out to a .txt-file."""

        try:
            # Initialise:
            fw = open(filename, mode='w')

            # Write all the immediate attributes:
            attributes = ['NAME', 'COMMENT', 'TYPE', 'AMOUNT_OF_NODES',
                          'AMOUNT_OF_EMERGENCY_NODES', 'NODE_ORDER', 'T',
                          'ADJACENCY_TYPE', 'ADJACENCY_FORMAT',
                          'RESPONSE_TIME_TYPE', 'RESPONSE_TIME_FORMAT']
            for arg in attributes:
                line = "%s: %s\n" % (arg, self.__dict__[arg])
                fw.write(line)

            # Write the NODE_COORD_SECTION:
            fw.write('NODE_COORD_SECTION\n')
            for i_v, v in enumerate(self.V):
                (x_coord, y_coord) = self.coord[v]
                line = ('%d\t%.2f\t%.2f\n') % (i_v+1, x_coord, y_coord)
                fw.write(line)

            # Write the ADJACENCY_SECTION:
            fw.write('ADJACENCY_SECTION\n')
            for i_u, u in enumerate(self.V):

                line = '%d\t' % (i_u+1)
                amount_on_line = 1
                remaining_neighbours = sorted(self.Vv[u])
                while remaining_neighbours:
                    neighb = remaining_neighbours[0]
                    remaining_neighbours.remove(neighb)
                    i_v = self.V.index(neighb)
                    if i_v > i_u:
                        line += '%d\t' % (i_v+1)
                        amount_on_line += 1
                        if amount_on_line >= 10:
                            line += '\n'
                            fw.write(line)
                            line = ''
                            amount_on_line = 0
                if line == '':
                    pass
                else:
                    line += '\n'
                    fw.write(line)
                if u != self.V[-1]:
                    fw.write('\n')

            # Write the AGENT_START_LOCATION_SECTION:
            fw.write('AGENT_START_LOCATION_SECTION\n')
            for i_a, a in enumerate(self.A):
                v = self.start[a]
                i_v = self.V.index(v)
                line = '%d\t%d\n' % (i_a+1, i_v+1)
                fw.write(line)

            # Write the AGENT_END_LOCATION_SECTION:
            fw.write('AGENT_END_LOCATION_SECTION\n')
            for i_a, a in enumerate(self.A):
                v = self.end[a]
                i_v = self.V.index(v)
                line = '%d\t%d\n' % (i_a+1, i_v+1)
                fw.write(line)

            # Write the JOB_LOCATION_SECTION:
            fw.write('JOB_LOCATION_SECTION\n')
            for i_j, j in enumerate(self.J):
                v = self.location[j]
                i_v = self.V.index(v)
                line = '%d\t%d\n' % (i_j+1, i_v+1)
                fw.write(line)

            # Write the JOB_DURATION_SECTION:
            fw.write('JOB_DURATION_SECTION\n')
            for i_j, j in enumerate(self.J):
                q = self.Q[j]
                line = '%d\t%d\n' % (i_j+1, q)
                fw.write(line)

            # Write the EMERGENCY_PROBABILITY_SECTION:
            fw.write('EMERGENCY_PROBABILITY_SECTION\n')
            skips = self.AMOUNT_OF_NODES - self.AMOUNT_OF_EMERGENCY_NODES
            for i_v, v in enumerate(self.VP):
                line = '%d\t%f\n' % (i_v+skips+1, self.P[v])
                fw.write(line)

            # Write the RESPONSE_TIME_SECTION (first the full part, then the
            # triangular part):
            fw.write('RESPONSE_TIME_SECTION\n')
            line = ''
            amount_on_line = 0
            # Full part:
            for i_u, u in enumerate(self.V[:skips]):
                for i_v, v in enumerate(self.VP):
                    line += '%.2f\t' % (self.C[u][v])
                    amount_on_line += 1
                    if amount_on_line >= 10:
                        line += '\n'
                        fw.write(line)
                        line = ''
                        amount_on_line = 0
            # Triangle part:
            for i_u, u in enumerate(self.VP):
                for i_v, v in enumerate(self.VP[:i_u]):
                    line += '%.2f\t' % (self.C[u][v])
                    amount_on_line += 1
                    if amount_on_line >= 10:
                        line += '\n'
                        fw.write(line)
                        line = ''
                        amount_on_line = 0
                        
            # Finish up:
            if line != '':
                line += '\n'
                fw.write(line)
            fw.write('EOF')

        except:
            print('Could not write instance to file.')
            print_exc()

        finally:
            fw.close()
            
    def plot(self):
        """Plot the instance, without solution overlaid."""
        
        try:
            
            import matplotlib.pyplot as plt
            
            # Abbreviate:
            A = self.A; J = self.J; V = self.V
            Q = self.Q; P = self.P
            pos = self.coord
            VP = self.VP; Vv = self.Vv
            start = self.start; location = self.location
            
            # Initialise graph and plot edges in dark gray:
            Gr = Graph()
            for i_u, u in enumerate(V):
                for v in Vv[u]:
                    i_v = V.index(v)
                    if i_v > i_u:
                        Gr.add_edge(u, v)
            draw_networkx_edges(Gr, pos, edge_color='#696969')

            # Categorise the nodes as follows:
            # - STARTS_V is each node where at least one agent starts
            # - JOBBED_V is the remainder which has jobs
            # - NORM_V is the remainder
            remainder = list(V)
            
            STARTS_V = list( self.start.values() )
            for a in A:
                v = start[a]
                if v in remainder:
                    remainder.remove(v)
            
            JOBBED_V = list()
            for j in J:
                v = location[j]
                if v in remainder:
                    JOBBED_V += [v]
                    remainder.remove(v)
                    
            NORM_V = remainder
            
            # Plot gray, round nodes for non-emergency, non-depot nodes:
            draw_networkx_nodes(Gr, pos, nodelist=NORM_V, node_size=75,
                                   node_color='#BDBDBD', node_shape='o')
                                   
            # Plot blue, square nodes for JOBBED_V:
            draw_networkx_nodes(Gr, pos, nodelist=JOBBED_V, node_size=75,
                                   node_color='b', node_shape='s')
                               
            # Plot a green pentagon for STARTS_V:
            draw_networkx_nodes(Gr, pos, nodelist=STARTS_V, node_size=75,
                                   node_color='g', node_shape='p')

            # Add overlay text to nodes according to the following rules:
            # - if v in DEPOT_V, add "|A| = ...";
            # - for each j, add "Q=..." to its node;
            # - if P[v] > 0, add "P = ...", rounded to two decimals.
            overlay = dict()
            
            for depot in STARTS_V:
                amount = len( [ a for a in A if start[a] == depot ] )
                overlay[depot] = "|A| = %d" % ( amount )
            
            for j in J:
                v = location[j]
                if v in overlay:
                    overlay[v] += ", Q=%d" % (Q[j])
                else:
                    overlay[v] = "Q=%d" % (Q[j])
                    
            for v in VP:
                if v in overlay:
                    overlay[v] += ", P=%.2f" % (P[v])
                else:
                    overlay[v] = "P=%.2f" % (P[v])

            # For each node, plot its text, if any:
            for v in overlay:
                x,y = pos[v]
                plt.text(x, y, overlay[v])
                
            plt.show()
              
        except:
            print('Something went wrong with plotting.')
            print_exc()

            
    def apply(self, solution_module):
        """
        Calls solution_module.apply on this instance, which should
        return a Solution. The solution_model should be geared specifically to
        this problem type.
        """
        
        return solution_module.apply(self)
    
    def solve(self):
        """
        Applies the default method for solving to optimality.
        """
        
        module = import_module(default_solve_method)
        return self.apply(module)

class Solution:
    """A solution of the Median Routing Problem."""
    
    def __init__(self, **kwargs):
        
        self.INSTANCE = None
        self.METHOD = None
        self.COST = None
        self.OPTIMAL = None
        self.x_SEQUENCES = None
        self.y_SEQUENCES = None
        self.z_ASSIGNMENTS = None
        
        for arg in kwargs:
            self.__dict__[arg] = kwargs[arg]
            
    def from_txt(filename, MRPinstance=None):
        """Read a solution from .txt."""
        
        raise NotImplementedError
        
    def to_txt(self, filename):
        """Write a solution to .txt."""
        
        try:
            fw = open(filename, 'w')
            
            # Write instance name:
            line = 'INSTANCE: %s\n' % str(self.INSTANCE.NAME)
            fw.write(line)
            
            # Write method comment:
            line = 'METHOD: %s\n' % str(self.METHOD)
            fw.write(line)
            
            # Write cost:
            line = 'COST: %f\n' % float(self.COST)
            fw.write(line)
            
            # Write whether or not optimal:
            line = 'GUARANTEED OPTIMAL: %s\n' % str(self.OPTIMAL)
            fw.write(line)
            
            # Write the x-sequences:
            fw.write('x-SEQUENCES\n')
            for a in self.x_SEQUENCES:
                line = ''
                for v in self.x_SEQUENCES[a]:
                    i_v = self.INSTANCE.V.index(v)
                    line += '%d ' % (i_v+1)
                line += '\n'
                fw.write(line)
                
            # Write the y-sequences:
            fw.write('y-SEQUENCES\n')
            for u in self.y_SEQUENCES:
                line = ''
                for v in self.y_SEQUENCES[u]:
                    i_v = self.INSTANCE.V.index(v)
                    line += '%d ' % (i_v+1)
                line += '\n'
                fw.write(line)
                
            # Write the z-assignments:
            fw.write('z-ASSIGNMENTS\n')
            for j in self.INSTANCE.J:
                (a, i_t) = self.z_ASSIGNMENTS[j]
                i_a = self.INSTANCE.A.index(a)
                line = '%d %d\n' % (i_a+1, i_t)
                fw.write(line)
                
        except:
            print('Could not write solution to file.')
            print_exc()
            
        finally:
            fw.close()
            
    def plot(self):
        """Plot this instance and overlay this solution."""
        
        try:
            import matplotlib.pyplot as plt
                
            # Abbreviate:
            MRPproblem = self.INSTANCE
            A = MRPproblem.A; J = MRPproblem.J; V = MRPproblem.V
            Q = MRPproblem.Q; P = MRPproblem.P
            pos = MRPproblem.coord
            VP = MRPproblem.VP; Vv = MRPproblem.Vv
            start = MRPproblem.start; location = MRPproblem.location
            T = ['t%d' % i for i in range(MRPproblem.T) ]
            Tm = T[:-1]
            x_SEQUENCES = self.x_SEQUENCES
            z_ASSIGNMENTS = self.z_ASSIGNMENTS
            location = MRPproblem.location
            
            # Figure out which time-steps people spend 'waiting' at a location,
            # by reporting waiting whenever an agent is still in the same place
            # at the next step, then overwriting 'not waiting' for the
            # processing of each job j:
            is_waiting = {a: {t: False for t in Tm} for a in A}
            for i_t, t in enumerate(Tm):
                for a in A:
                    if x_SEQUENCES[a][i_t] == x_SEQUENCES[a][ i_t+1 ]:
                        is_waiting[a][t] = True
            for j in J:
                q = MRPproblem.Q[j]
                (a,i_t) = z_ASSIGNMENTS[j]
                i_t = T.index(t)
                for tau in T[ i_t: i_t+q ]:
                    is_waiting[a][tau] = False
    
            # Categorise the nodes as follows:
            # - STARTS_V is each node where at least one agent starts
            # - JOBBED_V is the remainder which has jobs
            # - WAIT_V is the remainder with reasonless waiting
            # - NORM_V is the remainder
            remainder = list(V)
            
            STARTS_V = list( MRPproblem.start.values() )
            for a in A:
                v = start[a]
                if v in remainder:
                    remainder.remove(v)
                    
            JOBBED_V = list()
            for j in J:
                v = location[j]
                if v in remainder:
                    JOBBED_V += [v]
                    remainder.remove(v)
                    
            WAIT_V = list()
            for a in A:
                for i_t, t in enumerate(Tm):
                    if is_waiting[a][t]:
                        v = x_SEQUENCES[a][i_t]
                        if v in remainder:
                            WAIT_V += [v]
                            remainder.remove(v)
                    
            NORM_V = remainder
            
            # Categorise the edges as follows:
            # - GO_E = (u,v) if some agent goes from u to v
            # - NORM_E is the remainder if G[u][v], and not already (v,u) counted
            GO_E = list()
            for i_t, t in enumerate(Tm):
                for a in A:
                    u = x_SEQUENCES[a][ i_t ]
                    v = x_SEQUENCES[a][ i_t+1 ]
                    if u != v and not (u,v) in GO_E:
                        GO_E += [(u,v)]
                        
            NORM_E = list()
            for u in V:
                for v in Vv[u]:
                    if u != v and not (u,v) in GO_E and not (v,u) in GO_E + NORM_E:
                        NORM_E += [(u,v)]
            
            # Add overlay text to nodes according to the following rules:
            # - if v in STARTS_V, add "|A| = ...";
            # - for each j, add "Q=..." to its node;
            # - if P[v] > 0, add "P = ...", rounded to two decimals;
            # - looping over T then A, if someone waits at v, add "a... %d"
            overlay = dict()
            
            for depot in STARTS_V:
                amount = len( [ a for a in A if start[a] == depot ] )
                overlay[depot] = "|A| = %d" % ( amount )
                
            for j in J:
                v = location[j]
                if v in overlay:
                    overlay[v] += ", %Q=%d" % (Q[j])
                else:
                    overlay[v] = "Q=%d" % (Q[j])
                    
            for v in VP:
                if v in overlay:
                    overlay[v] += ", P=%.2f" % (P[v])
                else:
                    overlay[v] = "P=%.2f" % (P[v])
                    
            handled_waits = list()
            for i_t, t in enumerate(Tm):
                for a in A:
                    if is_waiting[a][t] and not (a,t) in handled_waits:
                        
                        # Count how long a waits, flagging waits as handled meanwhile:
                        wait = 0
                        for tau in Tm[i_t:]:
                            if is_waiting[a][tau]:
                                handled_waits += [(a,tau)]
                                wait += 1
                            else:
                                break
                        
                        # Add text:
                        v = x_SEQUENCES[a][i_t]
                        if v in overlay:
                            overlay[v] += ", %s... %d" % (a, wait)
                        else:
                            overlay[v] = "%s... %d" % (a, wait)
    
            # Initialise digraph and plot normal edges in dark gray:
            Gr = DiGraph()
            Gr.add_edges_from(NORM_E)
            Gr.add_edges_from(GO_E)
            draw_networkx_edges(Gr, pos, edgelist=NORM_E, 
                                edge_color='#696969', arrows=False)
                                
            # Plot the routes of the agents as light green arrows:
            draw_networkx_edges(Gr, pos, edgelist=GO_E,
                                edge_color='#00FF00', arrows=True)
            
            # Plot gray, round nodes for NORM_V:
            draw_networkx_nodes(Gr, pos, nodelist=NORM_V, node_size=75,
                                   node_color='#BDBDBD', node_shape='o')
                                   
            # Plot yellow, round nodes for WAIT_V:
            draw_networkx_nodes(Gr, pos, nodelist=WAIT_V, node_size=75,
                                    node_color='y', node_shape='o')
                                   
            # Plot blue, square nodes for JOBBED_V:
            draw_networkx_nodes(Gr, pos, nodelist=JOBBED_V, node_size=75,
                                   node_color='b', node_shape='s')
            
            # Plot a green pentagon for DEPOT_V:
            draw_networkx_nodes(Gr, pos, nodelist=STARTS_V, node_size=75,
                                   node_color='g', node_shape='p')
                                   
            # For each node, plot its text, if any:
            for v in overlay:
                x,y = pos[v]
                plt.text(x, y, overlay[v])
            
        except:
            print('Something went wrong with plotting.')
            print_exc()
    
