# -*- coding: utf-8 -*-
"""
Created on Mon Mar 11 08:35:20 2019

@author: Dylan
"""

cvrp_distance_decimals = 2

from copy import deepcopy
from pulp import LpBinary, LpContinuous, LpMinimize, LpProblem
from pulp import LpVariable, lpSum
from random import random
from random import choice as randomchoice

from MedianRoutingProblem import Instance

def eudist( xy_tuple1, xy_tuple2 ):
    """Short-hand for Euclidean distance between (x1,y1) and (x2,y2)."""
    sq_dist_x = pow( xy_tuple1[0] - xy_tuple2[0], 2 )
    sq_dist_y = pow( xy_tuple1[1] - xy_tuple2[1], 2 )
    dist = (sq_dist_x + sq_dist_y)**0.5
    return dist

def random_euclidean(**kwargs):
    """Creates and returns a random instance of the MRP on the Euclidean
    grid, where non-job/non-depot nodes have either large, medium or small
    emergency probability."""
    
    # Argument defaults:
    D = dict()
    D['size_A']         = 2
    D['size_J']         = 5
    D['size_V']         = 20
    D['T']              = 16
    D['scale']          = 1000
    D['G_chop']         = 300
    D['Q_range']        = range(0,3)
    D['probability_large_medium_small'] = [0.15, 0.3, 0.55]
    D['P_points_range_small'] = range(0,6)
    D['P_points_range_medium'] = range(10,51)
    D['P_points_range_large'] = range(100,301)
    D['attempts']       = 100
    
    # Override defaults:
    for arg in kwargs:
        D[arg] = kwargs[arg]
    
    # Randomisation may yield disconnected graphs, especially under
    # unfavourable parameters. Connected but infeasible instances are also 
    # undesirable. Only try a limited amount of times:
    attempt_i = 0
    while attempt_i < D['attempts']:
        
        print('Attempt %d...' % attempt_i)
        
        # Define a list of yet unclaimed coordinates on a grid:
        gridpoints = [(0,0)]
        for i in range(1, (D['scale']+1)**2 ):
            prev = gridpoints[i-1]
            if prev[0] == D['scale']:
                gridpoints += [ (0, prev[1]+1) ]
            else:
                gridpoints += [ (prev[0]+1, prev[1]) ]
        
        # Draw a random subset of gridpoints to determine V:
        chosen_gridpoints = list()
        for i in range( D['size_V'] ):
            xy = randomchoice(gridpoints)
            chosen_gridpoints += [xy]
            gridpoints.remove(xy)
            
        # Check if this yields a connected graph:
        Vv_temp = { xy1: {xy2 for xy2 in chosen_gridpoints 
                    if eudist(xy1, xy2) <= D['G_chop'] } 
                    for xy1 in chosen_gridpoints }
        connected = is_connected(Vv_temp, chosen_gridpoints)

        # If not so, try again:
        if not connected:
            attempt_i += 1
            print('Not connected.')
            continue
            
        # Name its centermost point depot='v0', move it to the front of the 
        # list, and denote V0 = [depot]:
        crd = dict()
        mid = ( 0.5*D['scale'], 0.5*D['scale'] )
        depot_xy = min(chosen_gridpoints, key = lambda xy: eudist( xy, mid ))
        chosen_gridpoints.remove(depot_xy)
        chosen_gridpoints = [depot_xy] + chosen_gridpoints
        crd['v0'] = depot_xy
        
        # On the next |J| chosen points, place jobs with randomly drawn
        # processing times. Denote these points V1:
        J = ['j%d' % (i) for i in range(D['size_J']) ]
        location = dict()
        Q = dict()
        for i_j, j in enumerate(J):
            Q[j] = randomchoice( D['Q_range'] )
            v = 'v%d' % (i_j+1)
            crd[v] = chosen_gridpoints[ i_j+1 ]
            location[j] = v
            
        # Check if these choices of depot, job locations and processing times
        # lead to a feasible instance:
        T = D['T']
        feasible = is_feasible( chosen_gridpoints, Vv_temp, J, Q, crd, T )
        
        # If not so, try again:
        if not feasible:
            attempt_i += 1
            print('Connected, but not feasible.')
            continue
            
        # If you've made it to here, a connected feasible instance was found:
        break
          
    if not connected or not feasible:
        print("Random generation failed: in %d attempts, no connected feasible instance was generated." % (D['attempts']))
        raise ValueError
        
    # Determine an ascending list of values for P by drawing an amount of 
    # 'incidents' for each non-job, non-depot node, then sorting and normalising:
    P_values = [ 0 for i in range( 1+D['size_J'] ) ]
    for i in range( 1 + D['size_J'], D['size_V'] ):
        
        # Determine whether this node has a 'large', 'medium' or 'small'
        # amount of emergencies, then draw accordingly:
        r = random()
        # Large:
        if r <= D['probability_large_medium_small'][0]:
            P_values += [ randomchoice( D['P_points_range_large'] ) ]
        # Small:
        elif r >= 1 - D['probability_large_medium_small'][-1]:
            P_values += [ randomchoice( D['P_points_range_small'] ) ]
        # Medium:
        else:
            P_values += [ randomchoice( D['P_points_range_medium'] ) ]
            
    # Sort and normalise:
    P_values = sorted(P_values)
    s = sum(P_values)
    P_values = [p/s for p in P_values]
        
    # Finish the labelling V and overlay the chosen grid points and P_values:
    V = ['v%d' % i for i in range( D['size_V'] )]
    P_temp = dict()
    for i, v in enumerate( V ):
        crd[v] = chosen_gridpoints[i]
        P_temp[v] = P_values[i]
    
    # Determine VP and clean up P:
    VP = [ v for v in V if P_temp[v] ]
    P = {v: P_temp[v] for v in VP}
    
    # Determine costs C_uv from truncated differences and adjacency Vv from 
    # untruncated distances:
    C = {u: {v: round(eudist(crd[u], crd[v]), 2) for v in VP} for u in V}
    Vv = {u: {v for v in V if eudist(crd[u], crd[v])<= D['G_chop']} for u in V}
    
    # Define the MRP instance, attach coordinates and return:
    A = ['a%d' % (i) for i in range(D['size_A'])]
    
    mrp = Instance()
    mrp.NAME = 'MRP-v%d-a%d-j%d-t%d-d%d' % (D['size_V'], D['size_A'],
                                            D['size_J'], D['T'], D['G_chop'])
    mrp.COMMENT = 'Randomly generated'
    mrp.TYPE = 'Median Routing Problem'
    mrp.AMOUNT_OF_NODES = len(V)
    mrp.AMOUNT_OF_EMERGENCY_NODES = len(VP)
    mrp.NODE_ORDER = 'NORMAL_THEN_EMERGENCY'
    mrp.ADJACENCY_TYPE = 'EXPLICIT'
    mrp.ADJACENCY_FORMAT = 'DECREASING_NEIGHBOUR_LIST'
    mrp.RESPONSE_TIME_TYPE = 'EXPLICIT'
    mrp.RESPONSE_TIME_FORMAT = 'FULL_THEN_LOWER_ROW'
    mrp.coord = crd
    mrp.start = {a: 'v0' for a in A}
    mrp.end = {a: 'v0' for a in A}
    mrp.location = location
    mrp.A = A
    mrp.J = J
    mrp.V = V
    mrp.VP = VP
    mrp.Vv = Vv
    mrp.T = T
    mrp.C = C
    mrp.Q = Q
    mrp.P = P
    
    return mrp

def given_euclidean(list_of_coord_tuples, list_of_probabilities,
                    list_of_job_coords, list_of_job_lengths, 
                    list_of_start_coords, list_of_end_coords, G_chop=1.0001, T=16):
    """Puts all nodes, jobs and agents on the given coordinates, then takes
    care of all Euclidean distances and adjacencies. Returns an instance."""
    
    # Create V, Pv, VP and crd:
    V = ['v_%f_%f' % coords for coords in list_of_coord_tuples]
    P = {v: list_of_probabilities[i_v] for i_v, v in enumerate(V) 
                                                if list_of_probabilities[i_v]}
    VP = [v for v in V if v in P]
    crd = {v: list_of_coord_tuples[i_v] for i_v, v in enumerate(V)}
    V = [v for v in V if not v in VP] + sorted(VP, key=lambda v: P[v])
    
    # Determine response times Cuv Euclideanly, then determine adjacencies by
    # checking the threshold G_chop:
    C = {u: {v: eudist(crd[u], crd[v]) for v in VP} for u in V}
    Vv = {u: {v for v in V if eudist(crd[u], crd[v]) <= G_chop} for u in V}
    
    # Add agents with start and end locations:
    A = ['a%d' % i for i in range(len(list_of_start_coords))]
    startloc = dict()
    endloc = dict()
    for i_a, a in enumerate(A):
        startloc[a] = 'v_%f_%f' %  list_of_start_coords[i_a]
        endloc[a] = 'v_%f_%f' %  list_of_end_coords[i_a]
        
    # Add jobs at right locations with right lengths:
    J = ['j%d' % i for i in range(len(list_of_job_coords))]
    Q = {j: list_of_job_lengths[i_j] for i_j, j in enumerate(J)}
    location = dict()
    for i_j, j in enumerate(J):
        location[j] = 'v_%f_%f' % list_of_job_coords[i_j]
    
    # Compile and return:
    mrp = Instance()
    mrp.NAME = 'Nameless'
    mrp.COMMENT = 'Created with benchmarkgenerator.given_euclidean().'
    mrp.TYPE = 'Median Routing Problem'
    mrp.AMOUNT_OF_NODES = len(V)
    mrp.AMOUNT_OF_EMERGENCY_NODES = len(VP)
    mrp.NODE_ORDER = 'NORMAL_THEN_EMERGENCY'
    mrp.ADJACENCY_TYPE = 'EXPLICIT'
    mrp.ADJACENCY_FORMAT = 'DECREASING_NEIGHBOUR_LIST'
    mrp.RESPONSE_TIME_TYPE = 'EXPLICIT'
    mrp.RESPONSE_TIME_FORMAT = 'FULL_THEN_LOWER_ROW'
    mrp.coord = crd
    mrp.start = startloc
    mrp.end = endloc
    mrp.location = location
    mrp.A = A
    mrp.J = J
    mrp.V = V
    mrp.VP = VP
    mrp.Vv = Vv
    mrp.T = T
    mrp.C = C
    mrp.Q = Q
    mrp.P = P
    
    return mrp

def is_connected(G, V):
    """Initiate a breadth-first search from the depot and see if all nodes
    are in its connected component. This is true iff the graph is connected.
    This version does not actually require a MRPinstance."""
    
    active_nodes = [ V[0] ]
    explored_nodes = set()
    while len(active_nodes) > 0:
        u = active_nodes[0]
        neighbours = G[u]
        for w in neighbours:
            if (not w in active_nodes) and (not w in explored_nodes):
                active_nodes += [w]
        explored_nodes.add(u)
        active_nodes.remove(u)
        if len(active_nodes) + len(explored_nodes) == len(V):
            break
            
    return len(active_nodes) + len(explored_nodes) == len(V)

def is_feasible( chosen_gridpoints, Vv_temp, J, Q, crd, T ):
    """Check if these choices for (single) depot, coordinates and job 
    processing times lead to a feasible instance, by computing the minimal 
    travel times between the jobs and depot and solving DVRP with objective 0."""
    
    # Determine shortest distances between depot and jobs:
    dJ = ['v%d' % i for i in range( len(J)+1 ) ]
    dist = { u: {v: 0 for v in dJ } for u in dJ }
    for i_u, u in enumerate(dJ):
        xy_u = chosen_gridpoints[i_u]
        for i_v, v in enumerate(dJ):
            if i_v > i_u:
                xy_v = chosen_gridpoints[i_v]
                d = minimal_dist( xy_u, xy_v, Vv_temp )
                dist[u][v] = dist[v][u] = d
            
    # Check for a feasible solution for DVRP by solving a MIP with objective 0:
    prob = LpProblem('feasibility', LpMinimize)
    
    # Variables:
    prob.w = w = LpVariable.dicts('w', (dJ,dJ), 0, 1, LpBinary )
    prob.f = f = LpVariable.dicts('f', (dJ[1:]), 0, T, LpContinuous )
    for j in dJ:
        w[j][j] = 0

    # Objective:
    prob += 0
    
    # Constraint (1):
    prob += lpSum( w['v0'][v] for v in dJ ) == lpSum( w[v]['v0'] for v in dJ ), "(1)"
    
    # Constraint (2):
    for v in dJ[1:]:
        prob += lpSum( w[u][v] for u in dJ ) == 1, "(2)_%s" % v
        
    # Constraint (3):
    for u in dJ[1:]:
        prob += lpSum( w[u][v] for v in dJ ) == 1, "(3)_%s" % u
        
    # Constraint (4):
    for i_u, u in enumerate(dJ):
        f_u = 0 if u == 'v0' else f[u]
        q = 0 if u == 'v0' else Q[ J[ i_u-1 ] ]
        for i_v, v in enumerate(dJ[1:]):
            j = J[i_v]
            prob += f[v] >= f_u + q + dist[u][v] - T*(1-w[u][v]), "(4)_%s_%s" % (u,v)
            
    # Constraint (5):
    for i_u, u in enumerate(dJ[1:]):
        j = J[i_u]
        prob += f[u] + Q[j] + dist[u]['v0'] <= T, "(5)_%s" % u
         
    # Solve and return if this succeeded:
    prob.solve()
    return (prob.status==1)
          
def minimal_dist( u, v, Vv_temp ):
    """Find out in how many steps minimally you can get from u to v."""
    
    active_nodes = [u]
    explored_nodes = set()
    steps_needed = {u: 0}
    
    # Width-first search:
    while active_nodes:
        
        w1 = active_nodes[0]
        active_nodes.remove(w1)
        explored_nodes.add(w1)
        
        for w2 in Vv_temp[w1]:
            
            if w2 == v:
                return steps_needed[w1] + 1
            
            if (not w2 in active_nodes) and (not w2 in explored_nodes):
                active_nodes += [w2]
                steps_needed[w2] = steps_needed[w1] + 1
    
    # If search terminated without finding v, something's wrong:
    raise ValueError

def unproductive_copy( MRPinstance ):
    """Returns an unproductive version of an MRP by keeping only |A|-1 jobs."""
    
    # Get a deepcopy:
    mrp2 = deepcopy( MRPinstance )
    
    # Keep only |A|-1 jobs and return:
    mrp2.J = mrp2.J[ : len(mrp2.A)-1 ]
    return mrp2

def dense_copy( MRPinstance, G_chop=600 ):
    """Returns a dense version of an MRP by redetermining Vv based on a new
    G_chop."""
    
    # Get a deepcopy:
    mrp2 = deepcopy( MRPinstance )
    
    # Reconfigure Vv and return:
    V = mrp2.V; crd = mrp2.coord
    for u in V:
        mrp2.Vv[u] = {v for v in V if eudist( crd[u], crd[v] ) <= G_chop }
        
    return mrp2
