# -*- coding: utf-8 -*-
"""
Created on Fri Jun  7 13:07:13 2019

@author: Dylan

This script reproduces the results of section I1 of Table 3 of the paper.
(That is: the number of feasible solutions, average gap, average time, worst
gap, and worst time found when applying the algorithms 
    opt, WAM, D, MD, DS, DA, MDS, MDA, DSA and MDSA
to all 50 instances of class I1.)
"""

from traceback import print_exc

try:
    
    import datetime; now = datetime.datetime.now
    from pandas import DataFrame
    
    from MedianRoutingProblem import Instance
    import MRP_gurobi
    import MRP_wam
    import MRP_mdsa_switches
    
    class_name = 'I1'
    class_size = 50
    
    # Prepare results dict:
    method_names = ['opt', 'WAM', 'D', 'MD', 'DS', 'DA', 
                    'MDS', 'MDA', 'DSA', 'MDSA']
    results = {'value': None, 'time': None}
    results['value'] = {method_name: dict() for method_name in method_names}
    results['time'] = {method_name: dict() for method_name in method_names}
    
    for i in range(class_size):
        
        # Read out problem data:
        filename = '../data/%s/%s_instance%d.txt' % (class_name, class_name, i)
        mrp = Instance.from_txt(filename)
        
        # Apply opt:
        succesful = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_gurobi)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['opt'][i] = mrs.COST
            results['time']['opt'][i] = toc-tic
            
        # Apply WAM:
        succesful = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_wam)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['WAM'][i] = mrs.COST
            results['time']['WAM'][i] = toc-tic
            
        # Apply D:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = False
        MRP_mdsa_switches.step_Sequence_is_on = False
        MRP_mdsa_switches.step_Agree_is_on = False
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['D'][i] = mrs.COST
            results['time']['D'][i] = toc-tic
            
        # Apply MD:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = True
        MRP_mdsa_switches.step_Sequence_is_on = False
        MRP_mdsa_switches.step_Agree_is_on = False
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['MD'][i] = mrs.COST
            results['time']['MD'][i] = toc-tic
            
        # Apply DS:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = False
        MRP_mdsa_switches.step_Sequence_is_on = True
        MRP_mdsa_switches.step_Agree_is_on = False
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['DS'][i] = mrs.COST
            results['time']['DS'][i] = toc-tic
            
        # Apply DA:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = False
        MRP_mdsa_switches.step_Sequence_is_on = False
        MRP_mdsa_switches.step_Agree_is_on = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['DA'][i] = mrs.COST
            results['time']['DA'][i] = toc-tic
            
        # Apply MDS:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = True
        MRP_mdsa_switches.step_Sequence_is_on = True
        MRP_mdsa_switches.step_Agree_is_on = False
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['MDS'][i] = mrs.COST
            results['time']['MDS'][i] = toc-tic
            
        # Apply MDA:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = True
        MRP_mdsa_switches.step_Sequence_is_on = False
        MRP_mdsa_switches.step_Agree_is_on = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['MDA'][i] = mrs.COST
            results['time']['MDA'][i] = toc-tic
            
        # Apply DSA:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = False
        MRP_mdsa_switches.step_Sequence_is_on = True
        MRP_mdsa_switches.step_Agree_is_on = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['DSA'][i] = mrs.COST
            results['time']['DSA'][i] = toc-tic
            
        # Apply MDSA:
        succesful = True
        MRP_mdsa_switches.step_Mediate_is_on = True
        MRP_mdsa_switches.step_Sequence_is_on = True
        MRP_mdsa_switches.step_Agree_is_on = True
        tic = now()
        try:
            mrs = mrp.apply(MRP_mdsa_switches)
        except:
            succesful = False
        toc = now()
        
        if succesful:
            results['value']['MDSA'][i] = mrs.COST
            results['time']['MDSA'][i] = toc-tic
            
    # Compile table:
    header_names = ['number of feasible solutions', 'average gap (%)', 
                    'average time (s)', 'worst gap (%)', 'worst time (s)']
    table = {header_name: {method_name: None for method_name in method_names} 
                                    for header_name in header_names}
    for method_name in method_names:
        amount_feasible = len(results['value'][method_name])
        table['number of feasible solutions'][method_name] = amount_feasible
        gap = [ (results['value'][method_name][i] / results['value']['opt'][i]
                - 1)*100 for i in range(class_size) ]
        time = [ (results['time'][method_name][key].seconds +
                  results['time'][method_name][key].microseconds/1000000)
                  for key in results['time'][method_name] ]
        table['average gap (%)'][method_name] = sum(gap)/len(gap)
        table['average time (s)'][method_name] = sum(time)/len(time)
        table['worst gap (%)'][method_name] = max(gap)
        table['worst time (s)'][method_name] = max(time)
        
    DF = DataFrame(table)
    print('Experiment succesful.')
    print('')
    print(DF)
    print('')
        
except:
    print_exc()
    
input('Press ENTER to close.')